//----------------------------------------------------------------------------
//
// File:        tms9918a-console.cpp
// Date:        23-Feb-1998
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 1994-2001 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <memory.h>
#include <stdio.h>
#include "common.hpp"
#include "compress.hpp"
#include "tms9900.hpp"
#include "tms9918a.hpp"
#include "tms9918a-console.hpp"
#include "screenio.hpp"

cConsoleTMS9918A::cConsoleTMS9918A ()
{
    m_Bias = 0;
}

cConsoleTMS9918A::~cConsoleTMS9918A ()
{
}

void cConsoleTMS9918A::Reset ()
{
    cTMS9918A::Reset ();

    m_Bias = 0;
    for ( int y = 1; y < 25; y++ ) PutXY ( 1, y, ( char * ) m_Memory, 32 );
}

void cConsoleTMS9918A::WriteData ( UCHAR data )
{
    if ((( m_Address & 0x3FFF ) >= ( UCHAR * ) m_ImageTable - m_Memory ) && (( m_Address & 0x3FFF ) < ( UCHAR * ) ( m_ImageTable + 1 ) - m_Memory )) {
        int imageTableOffset = ( UCHAR * ) m_ImageTable - m_Memory;
        int loc = ( m_Address & 0x3FFF ) - imageTableOffset;
        if (( loc >> 5 ) < 24 ) {
            UCHAR temp = ( UCHAR ) ( data - m_Bias );
            PutXY ( 1 + ( loc & 0x1F ), 1 + ( loc >> 5 ), ( char * ) &temp, 1 );
        }
    }

    cTMS9918A::WriteData ( data );
}

void cConsoleTMS9918A::WriteRegister ( int reg, UCHAR value )
{
    int x = 49 + ( reg % 4 ) * 9;
    int y = ( reg < 4 ) ? 10 : 11;
    GotoXY ( x, y );    outByte ( value );

    cTMS9918A::WriteRegister ( reg, value );
}

void cConsoleTMS9918A::SetBias ( UCHAR bias )
{
    if ( bias != m_Bias ) {
        m_Bias = bias;
        for ( unsigned i = ( UCHAR * ) m_ImageTable - m_Memory; i < sizeof ( sScreenImage ); i++ ) {
            UCHAR temp = ( UCHAR ) ( m_Memory [ i ] - m_Bias );
            if (( i >> 5 ) < 24 ) {
                PutXY ( 1 + ( i & 0x1F ), 1 + ( i >> 5 ), ( char * ) &temp, 1 );
            }
        }
    }
}

void cConsoleTMS9918A::LoadImage ( FILE *file )
{
    cTMS9918A::LoadImage ( file );

    for ( unsigned i = ( UCHAR * ) m_ImageTable - m_Memory; i < sizeof ( sScreenImage ); i++ ) {
        UCHAR temp = ( UCHAR ) ( m_Memory [ i ] - m_Bias );
        PutXY ( 1 + ( i & 0x1F ), 1 + ( i >> 5 ), ( char * ) &temp, 1 );
    }
}

