//----------------------------------------------------------------------------
//
// File:        dumpcpu.cpp
// Date:        23-Feb-1998
// Programmer:  Marc Rousseau
//
// Description:
//
// Copyright (c) 1994-2001 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <ctype.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.hpp"
#include "logger.hpp"
#include "cartridge.hpp"
#include "fileio.hpp"
#include "tms9900.hpp"

DBG_REGISTER ( __FILE__ );

#if defined ( DEBUG )
    LOG_FLAGS g_LogFlags;
#endif

extern "C" {
    UCHAR  CpuMemory [ 0x10000 ];
    sLookUp LookUp [ 16 ];
};

USHORT parity [ 256 ];

sTrapInfo TrapList  [ 16 ];

sOpCode   OpCodes [ 69 ] = {
  { "A   ", 0xA000, 0xF000, 1, ( USHORT ) -1, NULL, 14,   4 },	// 14
  { "AB  ", 0xB000, 0xF000, 1, ( USHORT ) -1, NULL, 14,   2 },	// 14
  { "ABS ", 0x0740, 0xFFC0, 6, ( USHORT ) -1, NULL, 12,   0 },	// 12/14
  { "AI  ", 0x0220, 0xFFE0, 8, ( USHORT ) -1, NULL, 14,  15 },	// 14
  { "ANDI", 0x0240, 0xFFE0, 8, ( USHORT ) -1, NULL, 14,  31 },	// 14
  { "B   ", 0x0440, 0xFFC0, 6, ( USHORT ) -1, NULL,  8, 126 },	// 8
  { "BL  ", 0x0680, 0xFFC0, 6, ( USHORT ) -1, NULL, 12,  36 },	// 12
  { "BLWP", 0x0400, 0xFFC0, 6, ( USHORT ) -1, NULL, 26,   0 },	// 26
  { "C   ", 0x8000, 0xF000, 1, ( USHORT ) -1, NULL, 14,  14 },	// 14
  { "CB  ", 0x9000, 0xF000, 1, ( USHORT ) -1, NULL, 14,   1 },	// 14
  { "CI  ", 0x0280, 0xFFE0, 8, ( USHORT ) -1, NULL, 14,  47 },	// 14
  { "CKOF", 0x03C0, 0xFFFF, 7, ( USHORT ) -1, NULL, 12,   0 },	// 12
  { "CKON", 0x03A0, 0xFFFF, 7, ( USHORT ) -1, NULL, 12,   0 },	// 12
  { "CLR ", 0x04C0, 0xFFC0, 6, ( USHORT ) -1, NULL, 10,  21 },	// 10
  { "COC ", 0x2000, 0xFC00, 3, ( USHORT ) -1, NULL, 14,   0 },	// 14
  { "CZC ", 0x2400, 0xFC00, 3, ( USHORT ) -1, NULL, 14,   0 },	// 14
  { "DEC ", 0x0600, 0xFFC0, 6, ( USHORT ) -1, NULL, 10,  60 },	// 10
  { "DECT", 0x0640, 0xFFC0, 6, ( USHORT ) -1, NULL, 10,   1 },	// 10
  { "DIV ", 0x3C00, 0xFC00, 9, ( USHORT ) -1, NULL, 16,   0 },	// 16/92-124
  { "IDLE", 0x0340, 0xFFFF, 7, ( USHORT ) -1, NULL, 12,   0 },	// 12 
  { "INC ", 0x0580, 0xFFC0, 6, ( USHORT ) -1, NULL, 10,  53 },	// 10
  { "INCT", 0x05C0, 0xFFC0, 6, ( USHORT ) -1, NULL, 10,   9 },	// 10
  { "INV ", 0x0540, 0xFFC0, 6, ( USHORT ) -1, NULL, 10,   2 },	// 10
  { "JEQ ", 0x1300, 0xFF00, 2, ( USHORT ) -1, NULL,  8,  39 },	// 8/10
  { "JGT ", 0x1500, 0xFF00, 2, ( USHORT ) -1, NULL,  8,  19 },	// 10
  { "JH  ", 0x1B00, 0xFF00, 2, ( USHORT ) -1, NULL,  8,   0 },	// 10
  { "JHE ", 0x1400, 0xFF00, 2, ( USHORT ) -1, NULL,  8,   4 },	// 10
  { "JL  ", 0x1A00, 0xFF00, 2, ( USHORT ) -1, NULL,  8,  18 },	// 10
  { "JLE ", 0x1200, 0xFF00, 2, ( USHORT ) -1, NULL,  8,   0 },	// 10
  { "JLT ", 0x1100, 0xFF00, 2, ( USHORT ) -1, NULL,  8,  70 },	// 10
  { "JMP ", 0x1000, 0xFF00, 2, ( USHORT ) -1, NULL,  8, 200 },	// 10
  { "JNC ", 0x1700, 0xFF00, 2, ( USHORT ) -1, NULL,  8,  27 },	// 10
  { "JNE ", 0x1600, 0xFF00, 2, ( USHORT ) -1, NULL,  8,  93 },	// 10
  { "JNO ", 0x1900, 0xFF00, 2, ( USHORT ) -1, NULL,  8,   0 },	// 10
  { "JOC ", 0x1800, 0xFF00, 2, ( USHORT ) -1, NULL,  8,   3 },	// 10
  { "JOP ", 0x1C00, 0xFF00, 2, ( USHORT ) -1, NULL,  8,   0 },	// 10
  { "LDCR", 0x3000, 0xFC00, 4, ( USHORT ) -1, NULL, 20,   1 },	// 20+2*bits
  { "LI  ", 0x0200, 0xFFE0, 8, ( USHORT ) -1, NULL, 12,  21 },	// 12
  { "LIMI", 0x0300, 0xFFE0, 8, ( USHORT ) -1, NULL, 16,  60 },	// 16
  { "LREX", 0x03E0, 0xFFFF, 7, ( USHORT ) -1, NULL, 12,   0 },	// 12
  { "LWPI", 0x02E0, 0xFFE0, 8, ( USHORT ) -1, NULL, 10,   1 },	// 10
  { "MOV ", 0xC000, 0xF000, 1, ( USHORT ) -1, NULL, 14, 144 },	// 14
  { "MOVB", 0xD000, 0xF000, 1, ( USHORT ) -1, NULL, 14, 420 },	// 14
  { "MPY ", 0x3800, 0xFC00, 9, ( USHORT ) -1, NULL, 52,   0 },	// 52
  { "NEG ", 0x0500, 0xFFC0, 6, ( USHORT ) -1, NULL, 12,   0 },	// 12
  { "ORI ", 0x0260, 0xFFE0, 8, ( USHORT ) -1, NULL, 14,  25 },	// 14
  { "RSET", 0x0360, 0xFFFF, 7, ( USHORT ) -1, NULL, 12,   0 },	// 12
  { "RTWP", 0x0380, 0xFFFF, 7, ( USHORT ) -1, NULL, 14,   0 },	// 14
  { "S   ", 0x6000, 0xF000, 1, ( USHORT ) -1, NULL, 14,   2 },	// 14
  { "SB  ", 0x7000, 0xF000, 1, ( USHORT ) -1, NULL, 14,   0 },	// 14
  { "SBO ", 0x1D00, 0xFF00, 2, ( USHORT ) -1, NULL, 12,   0 },	// 12
  { "SBZ ", 0x1E00, 0xFF00, 2, ( USHORT ) -1, NULL, 12,   0 },	// 12
  { "SETO", 0x0700, 0xFFC0, 6, ( USHORT ) -1, NULL, 10,   5 },	// 10
  { "SLA ", 0x0A00, 0xFF00, 5, ( USHORT ) -1, NULL, 12,  55 },	// 12+2*disp/20+2*disp
  { "SOC ", 0xE000, 0xF000, 1, ( USHORT ) -1, NULL, 14,   2 },	// 14
  { "SOCB", 0xF000, 0xF000, 1, ( USHORT ) -1, NULL, 14,   0 },	// 14
  { "SRA ", 0x0800, 0xFF00, 5, ( USHORT ) -1, NULL, 12,  16 },	// 12+2*disp/20+2*disp
  { "SRC ", 0x0B00, 0xFF00, 5, ( USHORT ) -1, NULL, 12,   0 },	// 12+2*disp/20+2*disp
  { "SRL ", 0x0900, 0xFF00, 5, ( USHORT ) -1, NULL, 12,  60 },	// 12+2*disp/20+2*disp
  { "STCR", 0x3400, 0xFC00, 4, ( USHORT ) -1, NULL, 42,   1 },	// 42/44/58/60
  { "STST", 0x02C0, 0xFFE0, 8, ( USHORT ) -1, NULL,  8,   3 },	// 8
  { "STWP", 0x02A0, 0xFFE0, 8, ( USHORT ) -1, NULL,  8,   0 },	// 8
  { "SWPB", 0x06C0, 0xFFC0, 6, ( USHORT ) -1, NULL, 10,  24 },	// 10
  { "SZC ", 0x4000, 0xF000, 1, ( USHORT ) -1, NULL, 14,   1 },	// 14
  { "SZCB", 0x5000, 0xF000, 1, ( USHORT ) -1, NULL, 14,  16 },	// 14
  { "TB  ", 0x1F00, 0xFF00, 2, ( USHORT ) -1, NULL, 12,   0 },	// 12
  { "X   ", 0x0480, 0xFFC0, 6, ( USHORT ) -1, NULL,  8,   0 },	// 8
  { "XOP ", 0x2C00, 0xFC00, 9, ( USHORT ) -1, NULL, 36,   0 },	// 36
  { "XOR ", 0x2800, 0xFC00, 3, ( USHORT ) -1, NULL, 14,   0 } 	// 14
};

#define BANK_SIZE	0x1000
#define BANK_MASK	0xF000

#define UNKNOWN		0x0000

#define DATA_BYTE	0x0001
#define DATA_WORD	0x0002
#define DATA_STRING	0x0004
#define DATA_TEXT	0x0008
#define DATA		0x000F

#define DATA_SINGLE	0x0010

#define PRESENT		0x0020

#define CODE		0x0100
#define CODE_START	0x0200

#define LABEL_DATA	0x0800
#define LABEL_JMP	0x1000
#define LABEL_B		0x2000
#define LABEL_BL	0x4000
#define LABEL_BLWP	0x8000
#define LABEL		0xF800

char   LastLabel  [5] = "A@";

char  *Labels     [ 0x10000 ];
USHORT Attributes [ 0x10000 ];
UCHAR  Arguments  [ 0x10000 ];

bool   CodeTraced;

extern "C" sLookUp LookUp [ 16 ];
extern "C" sOpCode OpCodes [ 69 ];

extern "C" UCHAR  CpuMemory  [ 0x10000 ];

extern USHORT DisassembleASM ( USHORT, UCHAR *, char * );

FILE *outFile;

int sortFunction ( const sOpCode *p1, const sOpCode *p2 )
{
    if ( p1->opCode < p2->opCode ) return -1;
    if ( p1->opCode > p2->opCode ) return 1;
    return 0;
}

void Init ()
{
    memset ( Attributes, 0, sizeof ( Attributes ));

    // Sort the OpCode table by OpCode
    qsort ( OpCodes, SIZE ( OpCodes ), sizeof ( OpCodes[0] ), ( QSORT_FUNC ) sortFunction );

    // Create the LookUp table using the high 4 bits of each OpCode
    unsigned x, last = 0;
    for ( x = 0; x < SIZE ( LookUp ) - 1; x++ ) {
        int index = last + 1;
        while ( x + 1 > ( unsigned ) ( OpCodes[index].opCode >> 12 )) index++;
        LookUp [x].opCode = &OpCodes [last];
        LookUp [x].size = index - last - 1;
        last = index;
    }
    LookUp [x].opCode = &OpCodes [last];
    LookUp [x].size = SIZE ( OpCodes ) - last - 1;
}

char *GetNextLabel ()
{
    FUNCTION_ENTRY ( NULL, "GetNextLabel", true );

    if ( LastLabel [1] == 'Z' ) {
        LastLabel [0]++;
        LastLabel [1] = 'A';
    } else {
        LastLabel [1]++;
    }
    return strdup ( LastLabel );
}

USHORT GetShort ( UCHAR *ptr )
{
    FUNCTION_ENTRY ( NULL, "GetShort", true );

    return ( ptr [0] << 8 ) | ptr [1];
}

struct sStackEntry {
    USHORT  address;
};

sStackEntry AddressStack [ 0x1000 ];
unsigned stackTop = 0;

bool PushAddress ( USHORT address, USHORT type, bool bottom = false )
{
    FUNCTION_ENTRY ( NULL, "PushAddress", true );

    if ( ! bottom ) Attributes [ address ] |= type;
    if ( stackTop >= SIZE ( AddressStack )) {
        fprintf ( stderr, "========= ** ERROR ** Address Stack Overflow ** ERROR ** =========\n" );
        exit ( -1 );
    }
    if ( Attributes [ address ] & CODE ) return true;
    if (( Attributes [ address ] & DATA ) && ( type != LABEL_BLWP )) ERROR ( "Pushed address marked as DATA" );

    sStackEntry *pEntry = AddressStack;

    if ( bottom ) {
        memmove ( AddressStack + 1, AddressStack, sizeof ( USHORT ) * stackTop++ );
    } else {
        pEntry += stackTop++;
    }

    pEntry->address = address;

    TRACE ( "--- pushing address >" << hex << address );

    return true;
}

bool IsStackEmpty ()
{
    FUNCTION_ENTRY ( NULL, "IsStackEmpty", true );

    return ( stackTop == 0 ) ? true : false;
}

bool PopAddress ( sStackEntry *entry )
{
    FUNCTION_ENTRY ( NULL, "PopAddress", true );

    if ( stackTop == 0 ) return false;

    *entry = AddressStack [ --stackTop ];

    TRACE ( "--- popping address >" << hex << entry->address );

    return true;
}

void InterpretHeader ( USHORT start )
{
    FUNCTION_ENTRY ( NULL, "InterpretHeader", true );

    if ( CpuMemory [ start ] != 0xAA ) return;

    USHORT *attr = Attributes + start;
    attr [ 0x0000 ] |= DATA_BYTE | DATA_SINGLE;
    attr [ 0x0001 ] |= DATA_BYTE | DATA_SINGLE;
    attr [ 0x0002 ] |= DATA_WORD | DATA_SINGLE;    attr [ 0x0003 ] |= DATA_WORD | DATA_SINGLE;
    attr [ 0x0004 ] |= DATA_WORD | DATA_SINGLE;    attr [ 0x0005 ] |= DATA_WORD | DATA_SINGLE;
    attr [ 0x0006 ] |= DATA_WORD | DATA_SINGLE;    attr [ 0x0007 ] |= DATA_WORD | DATA_SINGLE;
    attr [ 0x0008 ] |= DATA_WORD | DATA_SINGLE;    attr [ 0x0009 ] |= DATA_WORD | DATA_SINGLE;
    attr [ 0x000A ] |= DATA_WORD | DATA_SINGLE;    attr [ 0x000B ] |= DATA_WORD | DATA_SINGLE;
    attr [ 0x000C ] |= DATA_WORD | DATA_SINGLE;    attr [ 0x000D ] |= DATA_WORD | DATA_SINGLE;
    attr [ 0x000E ] |= DATA_WORD | DATA_SINGLE;    attr [ 0x000F ] |= DATA_WORD | DATA_SINGLE;

    // Interpret any headers we can find
    for ( int offset = start + 0x0004; offset < start + 0x000C; offset += 2 ) {
        USHORT next = GetShort ( CpuMemory + offset );
        while ( next ) {
            USHORT *attr = Attributes + next;
            USHORT address;
            *attr++ = DATA_WORD | DATA_SINGLE | LABEL;
            *attr++ = DATA_WORD | DATA_SINGLE;
            if ( isalpha ( CpuMemory [ next + 4 ])) {  	// Extended BASIC
                int len = CpuMemory [ next + 2 ] + 1;
                while ( len-- ) *attr++ = DATA_STRING;
                address = GetShort ( CpuMemory + next + 2 + len );
                *attr++ = DATA_WORD | DATA_SINGLE;
                *attr   = DATA_WORD | DATA_SINGLE;
            } else {					// TI BASIC
                address = GetShort ( CpuMemory + next + 2 );
                *attr++ = DATA_WORD | DATA_SINGLE;
                *attr   = DATA_WORD | DATA_SINGLE;
                int len = CpuMemory [ next + 4 ] + 1;
                while ( len-- ) *attr++ = DATA_STRING;
            }
            PushAddress ( address, LABEL_B );
            next = GetShort ( CpuMemory + next );
        }
    }

    // Interrupt Link
    USHORT address = GetShort ( CpuMemory + 0x000C );
    if ( address ) PushAddress ( address, LABEL_B );
}

void TraceCode ()
{
    FUNCTION_ENTRY ( NULL, "TraceCode", true );

    char buffer [ 80 ];

    sStackEntry top;

    while ( PopAddress ( &top ) == true ) {

        CodeTraced = true;

        USHORT address = top.address;

        if ( Attributes [ address ] & LABEL_BLWP ) {
            Attributes [ address ] |= DATA_WORD;
            Attributes [ address + 2 ] |= DATA_WORD;
            PushAddress ( GetShort ( CpuMemory + address + 2 ), 0 );
            continue;
        }

        for ( EVER ) {

            if (( Attributes [ address ] & PRESENT ) == 0 ) {
                WARNING ( "No code loaded @ >" << hex << address );
                break;
            }

            // If we've already been here, skip it
            if ( Attributes [ address ] & CODE ) {
                if (( Attributes [ address ] & CODE_START ) == 0 ) {
                    fprintf ( stderr, "Error - code doesn't begin with CODE_START!\n" );
                }
                break;
            }

            USHORT opcode = GetShort ( CpuMemory + address );

            // Mark all bytes in this instruction as CODE

            USHORT nextAddress = DisassembleASM ( address, &CpuMemory [ address ], buffer );

            TRACE ( "    " << buffer );

            int size = nextAddress - address;
            int badCode = 0;

            Attributes [ address ] |= CODE_START;

            for ( int i = 0; ( badCode == 0 ) && ( i < size ); i++ ) { 
                if ( Attributes [ address + i ] & CODE ) {
                    fprintf ( stderr, "Error parsing code - CODE attribute already set (%04X+%d)!\n", address, i );
                    badCode = i;
                    break;
                }
                if ( Attributes [ address + i ] & DATA ) {
                    fprintf ( stderr, "Error parsing code - DATA attribute set (%04X)!\n", address + i );
                    badCode = i + 1;
                    break;
                }
                if ( i && ( Attributes [ address + i ] & LABEL )) {
                    fprintf ( stderr, "Error parsing code - LABEL attribute set (%04X)!\n", address + i );
                    badCode = i + 1;
                    break;
                }
                Attributes [ address + i ] |= CODE;
            }
            if ( badCode ) {
                Attributes [ address ] &= ~CODE_START;
                for ( int i = 0; i < badCode; i++ ) {
                    Attributes [ address + i ] &= ~CODE;
                }
                break;
            }

            // Look for B instructions
            if (( opcode & 0xFFC0 ) == 0x0440 ) {
                if ( opcode == 0x0460 ) {
                    USHORT target = GetShort ( CpuMemory + address + 2 );
                    PushAddress ( target, LABEL_B );
                }
                break;
            }

            // Look for BL instructions
            if (( opcode & 0xFFC0 ) == 0x0680 ) {
                if ( opcode == 0x06A0 ) {
                    USHORT target = GetShort ( CpuMemory + address + 2 );
                    PushAddress ( target, LABEL_BL );
                    for ( int i = 0; i < Arguments [ target ]; i++ ) {
                        Attributes [ nextAddress + 2 * i ] |= DATA_WORD;
                        Attributes [ nextAddress + 2 * i + 1 ] |= DATA_WORD;
                    }
                    nextAddress += Arguments [ target ] * 2;
                }
            }

            // Look for BLWP instructions
            if (( opcode & 0xFFC0 ) == 0x0400 ) {
                USHORT target = GetShort ( CpuMemory + address + 2 );
                PushAddress ( target, LABEL_BLWP );
                for ( int i = 0; i < Arguments [ target ]; i++ ) {
                    Attributes [ nextAddress + 2 * i ] |= DATA_WORD;
                    Attributes [ nextAddress + 2 * i + 1 ] |= DATA_WORD;
                }
                nextAddress += Arguments [ target ] * 2;
            }

            // Look for jump instructions (except NOP)
            if (( opcode > 0x1000 ) && ( opcode < 0x1D00 )) {
                USHORT target = address + * ( char * ) &CpuMemory [ address + 1 ] * 2 + 2;
                PushAddress ( target, LABEL_JMP );
                if (( opcode & 0xFF00 ) == 0x1000 ) {
                    break;
                }
            }

            // Look for RTWP instructions
            if ( opcode == 0x0380 ) {
                break;
            }

            address = nextAddress;
        }
    }
}

USHORT DumpByte ( USHORT address, char *buffer )
{
    FUNCTION_ENTRY ( NULL, "DumpByte", true );

    USHORT limit = ( address & BANK_MASK ) + BANK_SIZE - 1;
    char *tempPtr = buffer + sprintf ( buffer, "BYTE " );
    for ( int i = 0; i < 8; i++ ) {
        if ( Attributes [ address ] & CODE ) {
            fprintf ( stderr, "**ERROR->%04X** %s DATA marked as %s\n", address, "Byte", "CODE" );
        }
        tempPtr += sprintf ( tempPtr, ">%02X", CpuMemory [ address ] );
        if ( ++address > limit ) break;
        if ( Attributes [ address - 1 ] & DATA_SINGLE ) break;
        if (( Attributes [ address ] != PRESENT ) && (( Attributes [ address ] & DATA_BYTE ) == 0 )) break;
        if (( Attributes [ address ] & PRESENT ) == 0 ) break;
        if ( Attributes [ address ] & LABEL ) break;
        if ( i < 7 ) *tempPtr++ = ',';
    }
    *tempPtr = '\0';
    return address;
}

USHORT DumpWord ( USHORT address, char *buffer )
{
    FUNCTION_ENTRY ( NULL, "DumpWord", true );

    USHORT limit = ( address & BANK_MASK ) + BANK_SIZE - 1;
    char *tempPtr = buffer + sprintf ( buffer, "DATA " );
    for ( int i = 0; i < 8; i += 2 ) {
        if ( Attributes [ address ] & CODE ) {
            fprintf ( stderr, "**ERROR->%04X** %s DATA marked as %s\n", address, "Word", "CODE" );
        }
        tempPtr += sprintf ( tempPtr, ">%04X", GetShort ( CpuMemory + address ));
        address += 2;
        if ( Attributes [ address - 2 ] & DATA_SINGLE ) break;
        if ( address > limit ) break;
        if (( Attributes [ address ] & DATA_WORD ) == 0 ) break;
        if (( Attributes [ address ] & PRESENT ) == 0 ) break;
        if ( Attributes [ address ] & LABEL ) break;
        if ( i < 6 ) *tempPtr++ = ',';
    }
    return address;
}

USHORT DumpString ( USHORT address, char *buffer )
{
    FUNCTION_ENTRY ( NULL, "DumpString", true );

    int length = CpuMemory [ address++ ];
    char *tempPtr = buffer + sprintf ( buffer, "STRI '" );
    for ( int i = 0; i < length; i++ ) {
        *tempPtr++ = CpuMemory [ address++ ];
    }
    *tempPtr++ = '\'';
    *tempPtr = '\0';
    return address;
}

USHORT DumpText ( USHORT address, char *buffer )
{
    FUNCTION_ENTRY ( NULL, "DumpText", true );

    USHORT limit = ( address & BANK_MASK ) + BANK_SIZE - 1;
    char *tempPtr = buffer + sprintf ( buffer, "BYTE '" );
    for ( int i = 0; i < 16; i++ ) {
        if ( Attributes [ address ] & CODE ) {
            ERROR ( "@ >" << hex << address << " - Text bytes marked as CODE" );
        }
        *tempPtr++ = CpuMemory [ address++ ];
        if ( address > limit ) break;
        if (( Attributes [ address ] & PRESENT ) == 0 ) break;
        if (( Attributes [ address ] & DATA_TEXT ) == 0 ) break;
        if ( Attributes [ address ] & LABEL ) break;
    }
    *tempPtr++ = '\'';
    *tempPtr = '\0';
    return address;
}

void AddLabels ( char *dstBuffer, char *srcBuffer )
{
    FUNCTION_ENTRY ( NULL, "AddLabels", true );

    do {
        char *ptr = strchr ( srcBuffer, '>' );
        if (( ptr == NULL ) || ( ! isxdigit ( ptr [2] ))) {
            strcpy ( dstBuffer, srcBuffer );
            return;
        }

        strncpy ( dstBuffer, srcBuffer, ptr - srcBuffer );
        dstBuffer += ptr - srcBuffer;

        int size = isxdigit ( ptr [3] ) ? 4 : 2;
        int address;
        sscanf ( ptr, ( size == 4 ) ? ">%04X" : ">%02X", &address );

        if ( Labels [ address ] ) {
            strcpy ( dstBuffer, Labels [ address ] );
            dstBuffer += strlen ( dstBuffer );
            ptr += size;
        } else if ( Attributes [ address ] & LABEL ) {
            dstBuffer += sprintf ( dstBuffer, "L%0*.*X", size, size, address );
            ptr += size;
        } else {
            *dstBuffer++ = '>';
        }
        srcBuffer = ptr + 1;
    } while ( *srcBuffer );

    *dstBuffer = '\0';
}

USHORT DumpCode ( USHORT address, char *buffer )
{
    FUNCTION_ENTRY ( NULL, "DumpCode", true );

    char tempBuffer [ 256 ];

    bool isCode = (( Attributes [ address ] & CODE ) != 0 ) ? true : false;

    if (( isCode == true ) && (( Attributes [ address ] & CODE_START ) == 0 )) {
        fprintf ( stderr, "**ERROR->%04X** Instruction doesn't begin with CODE_START\n", address );
    }

    USHORT nextAddress = DisassembleASM ( address, &CpuMemory [ address ], tempBuffer );
    bool ok = true;
    USHORT size = nextAddress - address;
    for ( int i = 0; i < size; i++ ) {
        if (( isCode == true ) && (( Attributes [ address + i ] & CODE ) == 0 )) {
            fprintf ( stderr, "**ERROR->%04X+%d** Instruction not marked as CODE\n", address, i );
            ok = false;
        }
        if ( i && ( Attributes [ address + i ] & CODE_START )) {
            fprintf ( stderr, "**ERROR->%04X+%d** Instruction marked with CODE_START\n", address, i );
            ok = false;
        }
        if ( Attributes [ address + i ] & DATA ) {
            fprintf ( stderr, "**ERROR->%04X+%d** Instruction marked as DATA\n", address, i );
        }
    }

    AddLabels ( buffer, tempBuffer + 5 );

    ok = ( buffer [1] == 'n' ) ? false : true;

    if ( ! ok ) {
        for ( int i = address; i < nextAddress; i++ ) {
            Attributes [ i ] = ( Attributes [ i ] & ~ ( CODE | CODE_START )) | DATA_WORD;
        }
        nextAddress = DumpWord ( address, buffer );
    }

    return nextAddress;
}

void DumpRam ( ULONG start, ULONG end )
{
    FUNCTION_ENTRY ( NULL, "DumpRam", true );

    fprintf ( outFile, "*\n" );
    fprintf ( outFile, "* Dump of ROM from %04X - %04X\n", start, end - 1 );
    fprintf ( outFile, "*\n" );

    ULONG address = start;
    ULONG lastAddress = start;

    printf ( "Dumping: %04X", address );

    bool aorg = true;

    while ( address < end ) {

        if (( address & 0xFFF0 ) != lastAddress ) {
            lastAddress = address & 0xFFF0;
            printf ( "%04X", lastAddress );
        }

        if ( aorg ) {
            fprintf ( outFile, "\t\tAORG >%04X\n\n", address );
            aorg = false;
        }

        char labelBuffer [ 32 ];
        if ( Labels [ address ] ) {
            strcpy ( labelBuffer, Labels [ address ] );
        } else {
            sprintf ( labelBuffer, "%c%04X", Attributes [ address ] & LABEL ? 'L' : ' ', address );
        }

        USHORT nextAddress = 0;
        char codeBuffer [ 256 ];
        if ( Attributes [ address ] & DATA ) {
            if ( Attributes [ address ] & DATA_STRING ) {
                nextAddress = DumpString (( USHORT ) address, codeBuffer );
            } else if ( Attributes [ address ] & DATA_TEXT ) {
                nextAddress = DumpText (( USHORT ) address, codeBuffer );
            } else if ( Attributes [ address ] & DATA_BYTE ) {
                nextAddress = DumpByte (( USHORT ) address, codeBuffer );
            } else if ( Attributes [ address ] & DATA_WORD ) {
                nextAddress = DumpWord (( USHORT ) address, codeBuffer );
            }
        } else if ( Attributes [ address ] & CODE ) {
            nextAddress = DumpCode (( USHORT ) address, codeBuffer );
        } else {
            if ( CodeTraced == true ) {
                nextAddress = DumpByte (( USHORT ) address, codeBuffer );
            } else {
                nextAddress = DumpCode (( USHORT ) address, codeBuffer );
            }
        }

        char dataBuffer [ 128 ];
        char *ptr = dataBuffer;

        if ( Attributes [ address ] & ( CODE | DATA_WORD )) {
            USHORT index = ( USHORT ) address;
            while ( index < nextAddress ) {
                ptr += sprintf ( ptr, ">%04X,", GetShort ( CpuMemory + index ));
                index += 2;
            }
            ptr [-1] = '\0';
        } else {
            USHORT index = ( USHORT ) address;
            while ( index < nextAddress ) {
                ptr += sprintf ( ptr, ">%02X,", CpuMemory [ index ]);
                index += 1;
            }
            ptr [-1] = '\0';
        }

        int tabs;
        tabs = 2 - strlen ( labelBuffer ) / 8;
        fprintf ( outFile, "%s%*.*s", labelBuffer, tabs, tabs, "\t\t\t\t\t" );
        tabs = 5 - strlen ( codeBuffer ) / 8;
        fprintf ( outFile, "%s%*.*s", codeBuffer, tabs, tabs, "\t\t\t\t\t" );
        tabs = 5 - ( strlen ( dataBuffer ) + 2 ) / 8;
        fprintf ( outFile, "* %s%*.*s", dataBuffer, tabs, tabs, "\t\t\t\t\t" );
        fprintf ( outFile, "* %04X\n", address );

        if (( ULONG ) nextAddress < address ) break;

        address = nextAddress;

        if (( Attributes [ address ] & PRESENT ) == 0 ) {
            fprintf ( outFile, "\n" );
            while ( address < end ) {
                address += 2;
                if ( Attributes [ address ] & PRESENT ) break;
            }
            aorg = true;
        }
    }

    printf ( "%04X\n", end );
}

bool g_LinePresent;
char g_LineBuffer [256];

void GetLine ( char *buffer, int size, FILE *file )
{
    FUNCTION_ENTRY ( NULL, "GetLine", true );

    if ( g_LinePresent ) {
        g_LinePresent = false;
        strcpy ( buffer, g_LineBuffer );
    } else {
        fgets ( buffer, size, file );
    }
}

void UnGetLine ( char *buffer )
{
    FUNCTION_ENTRY ( NULL, "UnGetLine", true );

    strcpy ( g_LineBuffer, buffer );
    g_LinePresent = true;
}

void ParseCode ( FILE *file )
{
    FUNCTION_ENTRY ( NULL, "ParseCode", true );

    for ( EVER ) {

        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
        if ( feof ( file )) break;

        if ( buffer [0] == '[' ) {
            UnGetLine ( buffer );
            break;
        }
        char *ptr = buffer;

        // Look for a label
        int symLength = 0;
        char symbol [256];
        while ( ! isspace ( *ptr )) {
            symbol [ symLength++ ] = *ptr++;
        }
        symbol [ symLength ] = 0;

        // Get the address
        while ( isspace ( *ptr )) {
            ptr++;
        }

        if ( *ptr == '\0' ) {
            continue;
        }

        int address;
        if ( sscanf ( ptr, "%04X", &address ) != 1 ) {
            fprintf ( stderr, "Error parsing line '%s'", buffer );
            break;
        }

        if ( symLength != 0 ) {
            Labels [ address ] = strdup ( symbol );
        }
        PushAddress ( address, 0, false );

    }
}

void ParseSubroutine ( FILE *file )
{
    FUNCTION_ENTRY ( NULL, "ParseSubroutine", true );

    for ( EVER ) {

        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
        if ( feof ( file )) break;

        if ( buffer [0] == '[' ) {
            UnGetLine ( buffer );
            break;
        }
        char *ptr = buffer;

        // Look for a label
        int symLength = 0;
        char symbol [256];
        while ( ! isspace ( *ptr )) {
            symbol [ symLength++ ] = *ptr++;
        }
        symbol [ symLength ] = 0;

        // Get the address
        while ( isspace ( *ptr )) {
            ptr++;
        }

        if ( *ptr == '\0' ) {
            continue;
        }

        int address, count;
        switch ( sscanf ( ptr, "%04X %d", &address, &count )) {
            case 1 :
                count = 0;
                break;
            case 2 :
                break;
            default :
                fprintf ( stderr, "Error parsing line '%s'", buffer );
                break;
        }

        if ( symLength != 0 ) {
            Labels [ address ] = strdup ( symbol );
        }
        PushAddress ( address, LABEL_BL, false );
        Arguments  [ address ] = count;

    }
}

void ParseBLWP ( FILE *file )
{
    FUNCTION_ENTRY ( NULL, "ParseBLWP", true );

    for ( EVER ) {

        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
        if ( feof ( file )) break;

        if ( buffer [0] == '[' ) {
            UnGetLine ( buffer );
            break;
        }
        char *ptr = buffer;

        // Look for a label
        int symLength = 0;
        char symbol [256];
        while ( ! isspace ( *ptr )) {
            symbol [ symLength++ ] = *ptr++;
        }
        symbol [ symLength ] = 0;

        // Get the address
        while ( isspace ( *ptr )) {
            ptr++;
        }

        if ( *ptr == '\0' ) {
            continue;
        }

        int address, count;
        switch ( sscanf ( ptr, "%04X %d", &address, &count )) {
            case 1 :
                count = 0;
                break;
            case 2 :
                break;
            default :
                fprintf ( stderr, "Error parsing line '%s'", buffer );
                break;
        }

        if ( symLength != 0 ) {
            Labels [ address ] = strdup ( symbol );
        }
        PushAddress ( address, LABEL_BLWP, false );
        Arguments  [ address ] = count;

    }
}

void ParseData ( FILE *file )
{
    FUNCTION_ENTRY ( NULL, "ParseData", true );

    for ( EVER ) {

        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
        if ( feof ( file )) break;

        if ( buffer [0] == '[' ) {
            UnGetLine ( buffer );
            break;
        }
        char *ptr = buffer;

        // Look for a label
        int symLength = 0;
        char symbol [256];
        while ( ! isspace ( *ptr )) {
            symbol [ symLength++ ] = *ptr++;
        }
        symbol [ symLength ] = 0;

        // Get the address, type, and length
        while ( isspace ( *ptr )) {
            ptr++;
        }

        if ( *ptr == '\0' ) {
            continue;
        }

        char type [256];
        int address, count;
        switch ( sscanf ( ptr, "%04X %s %04X", &address, type, &count )) {
            case 1 :
                strcpy ( type, "BYTE" );
            case 2 :
                count = 1;
                break;
            case 3 :
                break;
            default :
                fprintf ( stderr, "Error parsing line '%s'", buffer );
                break;
        }

        if ( symLength != 0 ) {
            Labels [ address ] = strdup ( symbol );
        }

        int attr = DATA_BYTE;
        if ( stricmp ( type, "BYTE" ) == 0 ) {
            attr = DATA_BYTE;
        }
        if ( stricmp ( type, "WORD" ) == 0 ) {
            attr = DATA_WORD;
            count *= 2;
        }
        if ( stricmp ( type, "STRING" ) == 0 ) {
            attr = DATA_STRING;
        }
        if ( stricmp ( type, "TEXT" ) == 0 ) {
            attr = DATA_TEXT;
        }

        Attributes [address] |= LABEL_DATA;
        for ( int i = 0; i < count; i++ ) {
            Attributes [address+i] |= attr;
        }
    }
}

void ParseTable ( FILE *file )
{
    FUNCTION_ENTRY ( NULL, "ParseTable", true );

    for ( EVER ) {

        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
        if ( feof ( file )) break;

        if ( buffer [0] == '[' ) {
            UnGetLine ( buffer );
            break;
        }
        char *ptr = buffer;

        // Look for a label
        int symLength = 0;
        char symbol [256];
        while ( ! isspace ( *ptr )) {
            symbol [ symLength++ ] = *ptr++;
        }
        symbol [ symLength ] = 0;

        // Get the address, type, and length
        while ( isspace ( *ptr )) {
            ptr++;
        }

        if ( *ptr == '\0' ) {
            continue;
        }

        int address, count;
        switch ( sscanf ( ptr, "%04X %04X", &address, &count )) {
            case 1 :
                count = 1;
                break;
            case 2 :
                break;
            default :
                fprintf ( stderr, "Error parsing line '%s'", buffer );
                break;
        }

        if ( symLength != 0 ) {
            Labels [ address ] = strdup ( symbol );
        }

        Attributes [address] |= LABEL_DATA;
        for ( int i = 0; i < 2 * count; i += 2 ) {
            Attributes [address+i] |= DATA_WORD;
            Attributes [address+i+1] |= DATA_WORD;
            USHORT code = GetShort ( CpuMemory + address + i );
            PushAddress ( code, LABEL_BL, false );
        }

    }
}

void ParseEQU ( FILE *file )
{
    FUNCTION_ENTRY ( NULL, "ParseEQU", true );

    for ( EVER ) {

        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
        if ( feof ( file )) break;

        if ( buffer [0] == '[' ) {
            UnGetLine ( buffer );
            break;
        }
        char *ptr = buffer;

        // Look for a label
        int symLength = 0;
        char symbol [256];
        while ( ! isspace ( *ptr )) {
            symbol [ symLength++ ] = *ptr++;
        }
        symbol [ symLength ] = 0;

        // Get the address
        while ( isspace ( *ptr )) {
            ptr++;
        }

        if ( *ptr == '\0' ) {
            continue;
        }

        int address;
        if ( sscanf ( ptr, "%04X", &address ) != 1 ) {
            fprintf ( stderr, "Error parsing line '%s'", buffer );
            break;
        }

        if ( symLength != 0 ) {
            Labels [ address ] = strdup ( symbol );
        }

    }
}

void ParseREF ( FILE *file )
{
    FUNCTION_ENTRY ( NULL, "ParseREF", true );

    for ( EVER ) {

        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
        if ( feof ( file )) break;

        if ( buffer [0] == '[' ) {
            UnGetLine ( buffer );
            break;
        }
    }
}

void ParseDEF ( FILE *file )
{
    FUNCTION_ENTRY ( NULL, "ParseDEF", true );

    for ( EVER ) {

        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
        if ( feof ( file )) break;

        if ( buffer [0] == '[' ) {
            UnGetLine ( buffer );
            break;
        }
    }
}

void ReadConfig ( const char *fileName )
{
    FUNCTION_ENTRY ( NULL, "ReadConfig", true );

    FILE *file = fopen ( fileName, "rt" );
    if ( file == NULL ) return;

    for ( EVER ) {

        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
        if ( feof ( file )) break;

        if ( buffer [0] != '[' ) continue;
        if ( strnicmp ( buffer + 1, "code", 4 ) == 0 ) {
          ParseCode ( file );
        }
        if ( strnicmp ( buffer + 1, "subroutine", 10 ) == 0 ) {
          ParseSubroutine ( file );
        }
        if ( strnicmp ( buffer + 1, "blwp", 4 ) == 0 ) {
          ParseBLWP ( file );
        }
        if ( strnicmp ( buffer + 1, "data", 4 ) == 0 ) {
          ParseData ( file );
        }
        if ( strnicmp ( buffer + 1, "table", 5 ) == 0 ) {
          ParseTable ( file );
        }
        if ( strnicmp ( buffer + 1, "equ", 3 ) == 0 ) {
          ParseEQU ( file );
        }
        if ( strnicmp ( buffer + 1, "ref", 3 ) == 0 ) {
          ParseREF ( file );
        }
        if ( strnicmp ( buffer + 1, "def", 3 ) == 0 ) {
          ParseDEF ( file );
        }
    }

    fclose ( file );
}

USHORT Hex2Value ( char *ptr )
{
    FUNCTION_ENTRY ( NULL, "Hex2Value", true );

    int digit;
    USHORT val = 0;

    digit = *ptr++ - '0';
    if ( digit > 9 ) digit -= 'A' - '0' - 10;
    val = ( val << 4 ) + digit;
    digit = *ptr++ - '0';
    if ( digit > 9 ) digit -= 'A' - '0' - 10;
    val = ( val << 4 ) + digit;
    digit = *ptr++ - '0';
    if ( digit > 9 ) digit -= 'A' - '0' - 10;
    val = ( val << 4 ) + digit;
    digit = *ptr - '0';
    if ( digit > 9 ) digit -= 'A' - '0' - 10;
    val = ( val << 4 ) + digit;

    return val;
}

/*
void LoadFile ( const char *fileName, ULONG *start, ULONG *end, ULONG *load )
{
    FUNCTION_ENTRY ( NULL, "LoadFile", true );

    *start = *end = *load = 0;

    FILE *file = fopen ( fileName, "rt" );
    if ( file == NULL ) return;

    ULONG low  = 0x00010000;
    ULONG high = 0x00000000;
    int addr = -1;

    do {
        char buffer [256];
        GetLine ( buffer, sizeof ( buffer ), file );
        char *ptr = buffer;
        bool done = false;
        while ( ! done ) {
            char tag = *ptr++;
            USHORT field = Hex2Value ( ptr );
            ptr += 4;
            switch ( tag ) {
                case '0' :	// Program Identification - length - ProgamID
                    ptr += 8;
                    break;
                case '1' :	// Entry Point Definition - Absolute
                    *load = field;
                    break;
                case '2' :	// Entry Point Definition - Relocatable
                    break;
                case '3' :	// External References - Absolute - Symbol
                case '4' :	// External References - Relocatable - Symbol
                case '5' :	// External Definitions - Absolute - Symbol
                case '6' :	// External Definitions - Relocatable - Symbol
                    ptr += 6;
                    break;
                case '7' :	// Checksum Indicator - Checksum
                    break;
                case '8' :	// Checksum Ignore - any value
                    break;
                case '9' :	// Load Address - Absolute
                    addr = field;
                    break;
                case 'A' :	// Load Address - Relocatable
                    break;
                case 'B' :	// Data - Absolute
                    CpuMemory [ addr ] = field >> 8;
                    CpuMemory [ addr+1] = field;
                    Attributes [ addr ] |= PRESENT;
                    addr += 2;
                    break;
                case 'C' :	// Data - Relocatable
                    break;
                case 'F' :	// End of Record
                case ':' :	// End of File
                default :
                    done = true;
                    break;
            }
            if ( addr != -1 ) {
                if ( addr > high ) high = addr;
                if ( addr < low ) low = addr;
            }
        }
    } while ( ! feof ( file ));

    *start = low;
    *end = high;

    fclose ( file );
}
*/

int main ( int, char *argv[] )
{
    FUNCTION_ENTRY ( NULL, "main", true );

    Init ();

    char baseName [ 256 ], fileName [ 256 ];
    strcpy ( baseName, argv [1] );
    char *_start = strrchr ( baseName, FILE_SEPERATOR );
    _start = ( _start == NULL ) ? baseName : _start + 1;
    char *ptr = strrchr ( _start, '.' );
    if ( ptr ) *ptr = '\0';

    sprintf ( fileName, "%s.%s", baseName, "ctg" );
    cCartridge ctg ( fileName );

    sprintf ( fileName, "%s.%s", _start, "asm" );
    outFile = fopen ( fileName, "wt" );
    if ( outFile == NULL ) {
        fprintf ( stderr, "Unable to open file '%s' for writing\n", fileName );
        return -1;
    }

    for ( unsigned i = 0; i < SIZE ( ctg.CpuMemory ); i++ ) {
        for ( int j = 0; j < ctg.CpuMemory[i].NumBanks; j++ ) {
            if ( ctg.CpuMemory[i].Bank[j].Type != MEMORY_ROM ) continue;
            UCHAR *ptr = ctg.CpuMemory[i].Bank[j].Data;
            memcpy (( char * ) &CpuMemory [ i * BANK_SIZE ], ( char * ) ptr, BANK_SIZE );
            for ( unsigned j = 0; j < BANK_SIZE; j++ ) {
                Attributes [ i * BANK_SIZE + j ] |= PRESENT;
            }
            break;
        }
    }
    
    sprintf ( fileName, "%s.%s", baseName, "dat" );
    ReadConfig ( fileName );

    for ( unsigned i = 0; i < SIZE ( ctg.CpuMemory ); i++ ) {
        for ( int j = 0; j < ctg.CpuMemory[i].NumBanks; j++ ) {
            if ( ctg.CpuMemory[i].Bank[j].Type != MEMORY_ROM ) continue;
            InterpretHeader ( i * BANK_SIZE );
            break;
        }
    }

    TraceCode ();

    fprintf ( outFile, "*\n" );
    fprintf ( outFile, "* Module: %s\n", ctg.Title ());
    fprintf ( outFile, "*\n" );

    for ( unsigned i = 0; i < SIZE ( ctg.CpuMemory ); i++ ) {
        for ( int j = 0; j < ctg.CpuMemory[i].NumBanks; j++ ) {
            if ( ctg.CpuMemory[i].Bank[j].Type != MEMORY_ROM ) continue;
            DumpRam ( i * BANK_SIZE, ( i + 1 ) * BANK_SIZE );
        }
    }

/*
    ULONG start, end, load;
    LoadFile ( baseName, &start, &end, &load );

    PushAddress ( load, LABEL_JMP );
    TraceCode ();

    DumpRam ( start, end );
*/

    fclose ( outFile );

    return 0;
}
